/*
** Module   : TempTableResultSet.java
** Abstract : A result set used to serialize temp table data and send it to a remote side. 
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130610 Created initial version.
*/

package com.goldencode.p2j.persist;

import java.util.*;

import org.hibernate.cfg.Configuration;
import org.hibernate.mapping.*;
import org.hibernate.mapping.List;
import org.hibernate.property.PropertyAccessor;

/**
 * Implements a method of serializing the metadata and the full data of a temp-table, specified
 * by a {@link Temporary temp-table DMO}.  The data is read fully in memory and sent to the remote
 * side row by row; the DMO metadata is used to extract the properties and their settings (type
 * and extent).  Note that each send row will have the values in the same order as the DMO's
 * properties. 
 */
public class TempTableResultSet
extends TableResultSet
{
   /**
    * The DMO associated with the temp table; this will be used to extract the DMO metadata and 
    * the table data.
    */
   private Temporary dmo;
   
   /** Iterator over the DMO's properties. */
   private Iterator<String> propIter;
   
   /** Iterator over the temp-table's data. */
   private Iterator<Object[]> rowIter;
   
   /**
    * Create a new instance of this result set and associate it with the metadata and data for
    * the specified temp DMO.
    * 
    * @param    dmo
    *           DMO instance returned by a previous call to {@link TemporaryBuffer#define} on the 
    *           source temporary buffer.
    * @param    input
    *           Flag indicating this table is sent in INPUT or INPUT-OUTPUT mode.
    * @param    output
    *           Flag indicating this table is sent in OUTPUT or INPUT-OUTPUT mode.
    * @param    append
    *           Flag indicating this table is sent in APPEND mode.
    */
   public TempTableResultSet(Temporary dmo, boolean input, boolean output, boolean append)
   {
      super(input, output, append);
      this.dmo = dmo;
      init();
   }
   
   /**
    * Check if there are more properties for this DMO.
    * 
    * @return   See above.
    */
   public boolean hasMoreProperties()
   {
      return propIter.hasNext();
   }
   
   /**
    * Returns the next property of this DMO.
    * 
    * @return   See above.
    */
   public String nextProperty()
   {
      return propIter.next();
   }
   
   /**
    * Check if there are more rows for this table.
    * 
    * @return   See above.
    */
   public boolean hasMoreRows()
   {
      return rowIter.hasNext();
   }
   
   /**
    * Returns the data for the next row.
    * 
    * @return    See above.
    */
   public Object[] nextRow()
   {
      return rowIter.next();
   }
   
   /**
    * Initialize this result set, by reading the entire temp-table in memory.
    * 
    * TODO: improve this (i.e. use an open cursor or some other alternative to not read the data
    * in memory)
    */
   private void init()
   {
      TemporaryBuffer buffer = (TemporaryBuffer) ((BufferReference) dmo).buffer();
      Class<?> srcClass = buffer.getDMOImplementationClass();

      PersistentClass cls = DatabaseManager.getClassMapping(DatabaseManager.TEMP_TABLE_DB,  
                                                            srcClass);
      
      java.util.List<String> props = new ArrayList<String>();
      types = new HashMap<String, String>();
      extents = new HashMap<String, Integer>();
      
      Iterator<Property> iter = cls.getPropertyIterator();
      while (iter.hasNext())
      {
         Property prop = iter.next();
         Value value = prop.getValue();
         
         // skip properties that are not accessible via the DMO's API.
         PropertyAccessor pa = prop.getPropertyAccessor(buffer.getDMOImplementationClass());
         if (pa instanceof NopPropertyAccessor)
         {
            continue;
         }

         if (value instanceof List)
         {
            List list = (List) value;
            Value element = list.getElement();
            if (element instanceof Component)
            {
               Component comp = (Component) element;
               
               // compute the number of entries in this component
               String compName = comp.getComponentClassName();
               compName = compName.substring(compName.indexOf("$Composite"));
               compName = compName.substring("$Composite".length());
               int extent = Integer.parseInt(compName);

               Iterator<Property> itrc = comp.getPropertyIterator();
               while (itrc.hasNext())
               {
                  Property compProp = (Property) itrc.next();
                  SimpleValue compVal = (SimpleValue) compProp.getValue();
                  
                  String propName = compProp.getName();
                  props.add(propName);
                  types.put(propName, compVal.getTypeName());
                  extents.put(propName, extent);
               }
            }
         }
         else
         {
            String propName = prop.getName();
            SimpleValue val = (SimpleValue) value;

            props.add(propName);
            types.put(propName, val.getTypeName());
         }
      }

      rowIter = TemporaryBuffer.readAllRows(dmo, props).iterator();
      propIter = props.iterator();
   }
}
