/*
** Module   : Account.java
** Abstract : Abstract class that represents a security account.
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description--------------------------------
** 001 NVS 20050307   @20193 Created the initial version.
** 002 NVS 20050311   @20292 Added abstract method getAlias().
** 003 NVS 20060412   @25491 Added new constant, ACC_ALL.
** 004 GES 20081104   @40338 Moved common code here from subclasses.
** 005 NVS 20090603   @42589 Added new member, enabled, and related methods 
**                           isEnabled(), setEnabled().
** 006 SVL 20100920          The class was made public.
** 007 CA  20130529          Added appserver support.
*/ 

package com.goldencode.p2j.security;

/**
 * Abstract class representing a security account. 
 */
public abstract class Account
{
   /** Account type - process. */
   public static final int ACC_PROCESS = 0;

   /** Account type - user. */
   public static final int ACC_USER = 1;
   
   /** Aaccount type - group. */
   public static final int ACC_GROUP = 2;

   /** Account selection type - all. */
   public static final int ACC_ALL = 3;

   /** The subject ID this account represents. */
   private String subjectId = null;

   /** Text description of this group. */
   private String description = null;

   /** Truststore alias of the associated certificate entry. */
   private String alias = null;

   /** Enabled status of this account. */
   private boolean enabled = true;
   
   /** The name of the app server. May be <code>null</code>. */
   private String appServer = null;
   
   /**
    * Constructor.
    *
    * @param    subjectId
    *           The account's subject ID (also known as account name).
    * @param    enabled
    *           <code>true</code> if this account is enabled
    * @param    description
    *           The verbose description of the account.
    * @param    alias
    *           The certificate entry name for this account in the truststore 
    *           This may be <code>null</code>.
    * @param    appServer
    *           The name of the app server. May be <code>null</code>.
    */
   Account(String subjectId, boolean enabled, String description, String alias, String appServer)
   {
      this.subjectId   = subjectId;
      this.description = description;
      this.alias       = alias;
      this.enabled     = enabled;
      this.appServer   = appServer;
   }
   
   /**
    * Constructor.
    *
    * @param    subjectId
    *           The account's subject ID (also known as account name).
    * @param    enabled
    *           <code>true</code> if this account is enabled
    * @param    description
    *           The verbose description of the account.
    * @param    alias
    *           The certificate entry name for this account in the
    *           truststore.  This may be <code>null</code>.
    */
   Account(String subjectId, boolean enabled, String description, String alias)
   {
      this(subjectId, enabled, description, alias, null);
   }

   /**
    * Gets the enabled status of this account.
    * Only the user and process accounts can be disabled. Disabled accounts
    * can't be authenticated successfully.
    *
    * @return   <code>true</code> if the account is enabled
    */
   boolean isEnabled()
   {
      return enabled;
   }
   
   /**
    * Changes the enabled status of this account.
    *
    * @param enabled
    *        <code>true</code> if the account is enabled
    */
   void setEnabled(boolean enabled)
   {
      this.enabled = enabled;
   }
   
   /**
    * Gets the subject ID for this account.
    *
    * @return   The subject ID.
    */
   String getSubjectId()
   {
      return subjectId;
   }
   
   /**
    * Gets the verbose description for this account.
    *
    * @return   The description.
    */
   String getDescription()
   {
      return description;
   }
   
   /**
    * Gets the certificate entry alias in the trust store for this account.
    *
    * @return   The alias if defined or <code>null</code>.
    */
   String getAlias()
   {
      return alias;
   }
   
   /**
    * Get the appserver name associated with this account.
    * 
    * @return    The {@link #appServer} name.
    */
   String getAppServer()
   {
      return appServer;
   }
   
   /**
    * Gets the account type of this account.
    *
    * @return   The account type as one of the <code>ACC_*</code> constants.
    */
   abstract int getAccountType();
}

